
import re
import typing

__all__ = (
    'get_language',
    'guess_file_traits',
    'LANGUAGES'
)


LANGUAGES = sorted([
    'as',
    '1c',
    'abnf',
    'accesslog',
    'actionscript',
    'ada',
    'ado',
    'adoc',
    'apache',
    'apacheconf',
    'applescript',
    'arduino',
    'arm',
    'armasm',
    'asciidoc',
    'aspectj',
    'atom',
    'autohotkey',
    'autoit',
    'avrasm',
    'awk',
    'axapta',
    'bash',
    'basic',
    'bat',
    'bf',
    'bind',
    'bnf',
    'brainfuck',
    'c',
    'c++',
    'cal',
    'capnp',
    'capnproto',
    'cc',
    'ceylon',
    'clean',
    'clj',
    'clojure-repl',
    'clojure',
    'cls',
    'cmake.in',
    'cmake',
    'cmd',
    'coffee',
    'coffeescript',
    'console',
    'coq',
    'cos',
    'cpp',
    'cr',
    'craftcms',
    'crm',
    'crmsh',
    'crystal',
    'cs',
    'csharp',
    'cson',
    'csp',
    'css',
    'd',
    'dart',
    'dcl',
    'delphi',
    'dfm',
    'diff',
    'django',
    'dns',
    'do',
    'docker',
    'dockerfile',
    'dos',
    'dpr',
    'dsconfig',
    'dst',
    'dts',
    'dust',
    'ebnf',
    'elixir',
    'elm',
    'erb',
    'erl',
    'erlang-repl',
    'erlang',
    'excel',
    'f90',
    'f95',
    'feature',
    'fix',
    'flix',
    'fortran',
    'freepascal',
    'fs',
    'fsharp',
    'gams',
    'gauss',
    'gcode',
    'gemspec',
    'gherkin',
    'glsl',
    'gms',
    'go',
    'golang',
    'golo',
    'gradle',
    'graph',
    'groovy',
    'gss',
    'gyp',
    'h',
    'h++',
    'haml',
    'handlebars',
    'haskell',
    'haxe',
    'hbs',
    'hpp',
    'hs',
    'hsp',
    'html.handlebars',
    'html.hbs',
    'html',
    'htmlbars',
    'http',
    'https',
    'hx',
    'hy',
    'hylang',
    'i7',
    'iced',
    'icl',
    'inform7',
    'ini',
    'instances',
    'irb',
    'irpf90',
    'java',
    'javascript',
    'jboss-cli',
    'jinja',
    'js',
    'json',
    'jsp',
    'jsx',
    'julia',
    'k',
    'kdb',
    'kotlin',
    'lasso',
    'lassoscript',
    'lazarus',
    'ldif',
    'leaf',
    'less',
    'lfm',
    'lisp',
    'livecodeserver',
    'livescript',
    'llvm',
    'lpr',
    'ls',
    'lsl',
    'lua',
    'm',
    'mak',
    'makefile',
    'markdown',
    'mathematica',
    'matlab',
    'maxima',
    'md',
    'mel',
    'mercury',
    'mips',
    'mipsasm',
    'mizar',
    'mk',
    'mkd',
    'mkdown',
    'ml',
    'mm',
    'mma',
    'mojolicious',
    'monkey',
    'moo',
    'moon',
    'moonscript',
    'n1ql',
    'nc',
    'nginx',
    'nginxconf',
    'nim',
    'nimrod',
    'nix',
    'nixos',
    'nsis',
    'obj-c',
    'objc',
    'objectivec',
    'ocaml',
    'openscad',
    'osascript',
    'oxygene',
    'p21',
    'parser3',
    'pas',
    'pascal',
    'patch',
    'pb',
    'pbi',
    'pcmk',
    'perl',
    'pf.conf',
    'pf',
    'php',
    'php3',
    'php4',
    'php5',
    'php6',
    'pl',
    'plist',
    'pm',
    'podspec',
    'pony',
    'powershell',
    'pp',
    'processing',
    'profile',
    'prolog',
    'protobuf',
    'ps',
    'puppet',
    'purebasic',
    'py',
    'python',
    'q',
    'qml',
    'qt',
    'r',
    'rb',
    'rib',
    'roboconf',
    'rs',
    'rsl',
    'rss',
    'ruby',
    'ruleslanguage',
    'rust',
    'scad',
    'scala',
    'scheme',
    'sci',
    'scilab',
    'scss',
    'sh',
    'shell',
    'smali',
    'smalltalk',
    'sml',
    'sqf',
    'sql',
    'st',
    'stan',
    'stata',
    'step',
    'step21',
    'stp',
    'styl',
    'stylus',
    'subunit',
    'sv',
    'svh',
    'swift',
    'taggerscript',
    'tao',
    'tap',
    'tcl',
    'tex',
    'thor',
    'thrift',
    'tk',
    'toml',
    'tp',
    'ts',
    'twig',
    'typescript',
    'v',
    'vala',
    'vb',
    'vbnet',
    'vbs',
    'vbscript-html',
    'vbscript',
    'verilog',
    'vhdl',
    'vim',
    'wildfly-cli',
    'x86asm',
    'xhtml',
    'xjb',
    'xl',
    'xls',
    'xlsx',
    'xml',
    'xpath',
    'xq',
    'xquery',
    'xsd',
    'xsl',
    'yaml',
    'yml',
    'zep',
    'zephir',
    'zone',
    'zsh'
], key=len, reverse=True)


def get_language(query: str) -> str:
    """Tries to work out the highlight.js language of a given file name or
    shebang. Returns an empty string if none match.
    """
    query = query.lower()
    for language in LANGUAGES:
        if query.endswith(language):
            return language
    return ''


ENCODING_REGEX = re.compile(br'coding[=:]\s*([-\w.]+)')


def guess_file_traits(data: bytes) -> typing.Tuple[str, str, typing.Optional[str]]:
    """
    Given the content of a file, attempts to guess its encoding and language.
    Returns as a tuple of (content, encoding, language),
    where language may be None.
    Raises UnicodeDecodeError if the encoding cannot be guessed.
    """

    try:
        content = data.decode('utf-8')
        encoding = 'utf-8'
    except UnicodeDecodeError as exc:
        # This file isn't UTF-8.

        #  By Python and text-editor convention,
        # there may be a hint as to what the actual encoding is
        # near the start of the file.

        encoding_match = ENCODING_REGEX.search(data[:128])

        if encoding_match:
            encoding = encoding_match.group(1)
        else:
            raise exc

        try:
            encoding = encoding.decode('utf-8')
            content = data.decode(encoding)
        except UnicodeDecodeError as exc2:
            raise exc2 from exc

    language = None

    if content.startswith('#!') and '\n' in content:
        language = get_language(content[:content.find('\n')]) or language

    return content, encoding, language


# -*- coding: utf-8 -*-

"""
jishaku.repl.walkers
~~~~~~~~~~~~~~~~~~~~

AST walkers for code transformation and analysis.

:copyright: (c) 2021 Devon (Gorialis) R
:license: MIT, see LICENSE for more details.

"""

import ast

# pylint: disable=no-self-use,invalid-name,missing-docstring


class KeywordTransformer(ast.NodeTransformer):
    """
    This transformer:
    - Converts return-with-value into yield & return
    - Converts bare deletes into conditional global pops
    """

    def visit_FunctionDef(self, node):
        # Do not affect nested function definitions
        return node

    def visit_AsyncFunctionDef(self, node):
        # Do not affect nested async function definitions
        return node

    def visit_ClassDef(self, node):
        # Do not affect nested class definitions
        return node

    def visit_Return(self, node):
        # Do not modify valueless returns
        if node.value is None:
            return node

        # Otherwise, replace the return with a yield & valueless return
        return ast.If(
            test=ast.NameConstant(
                value=True,  # if True; aka unconditional, will be optimized out
                lineno=node.lineno,
                col_offset=node.col_offset,
            ),
            body=[
                # yield the value to be returned
                ast.Expr(
                    value=ast.Yield(value=node.value, lineno=node.lineno, col_offset=node.col_offset),
                    lineno=node.lineno,
                    col_offset=node.col_offset,
                ),
                # return valuelessly
                ast.Return(value=None, lineno=node.lineno, col_offset=node.col_offset),
            ],
            orelse=[],
            lineno=node.lineno,
            col_offset=node.col_offset,
        )

    def visit_Delete(self, node):
        """
        This converter replaces bare deletes with conditional global pops.

        It is roughly equivalent to transforming:

        .. code:: python

            del foobar

        into:

        .. code:: python

            if 'foobar' in globals():
                globals().pop('foobar')
            else:
                del foobar

        This thus makes deletions in retain mode work more-or-less as intended.
        """

        return ast.If(
            test=ast.NameConstant(
                value=True,  # if True; aka unconditional, will be optimized out
                lineno=node.lineno,
                col_offset=node.col_offset,
            ),
            body=[
                ast.If(
                    # if 'x' in globals():
                    test=ast.Compare(
                        # 'x'
                        left=ast.Str(s=target.id, lineno=node.lineno, col_offset=node.col_offset),
                        ops=[
                            # in
                            ast.In(lineno=node.lineno, col_offset=node.col_offset)
                        ],
                        comparators=[
                            # globals()
                            self.globals_call(node)
                        ],
                        lineno=node.lineno,
                        col_offset=node.col_offset,
                    ),
                    body=[
                        ast.Expr(
                            # globals().pop('x')
                            value=ast.Call(
                                # globals().pop
                                func=ast.Attribute(
                                    value=self.globals_call(node),
                                    attr="pop",
                                    ctx=ast.Load(),
                                    lineno=node.lineno,
                                    col_offset=node.col_offset,
                                ),
                                args=[
                                    # 'x'
                                    ast.Str(
                                        s=target.id,
                                        lineno=node.lineno,
                                        col_offset=node.col_offset,
                                    )
                                ],
                                keywords=[],
                                lineno=node.lineno,
                                col_offset=node.col_offset,
                            ),
                            lineno=node.lineno,
                            col_offset=node.col_offset,
                        )
                    ],
                    # else:
                    orelse=[
                        # del x
                        ast.Delete(
                            targets=[target],
                            lineno=node.lineno,
                            col_offset=node.col_offset,
                        )
                    ],
                    lineno=node.lineno,
                    col_offset=node.col_offset,
                )
                if isinstance(target, ast.Name)
                else ast.Delete(targets=[target], lineno=node.lineno, col_offset=node.col_offset)
                # for each target to be deleted, e.g. `del {x}, {y}, {z}`
                for target in node.targets
            ],
            orelse=[],
            lineno=node.lineno,
            col_offset=node.col_offset,
        )

    def globals_call(self, node):
        """
        Creates an AST node that calls globals().
        """

        return ast.Call(
            func=ast.Name(
                id="globals",
                ctx=ast.Load(),
                lineno=node.lineno,
                col_offset=node.col_offset,
            ),
            args=[],
            keywords=[],
            lineno=node.lineno,
            col_offset=node.col_offset,
        )




import asyncio
import subprocess
import sys
import traceback
import typing
from types import TracebackType

import discord
from discord.ext import commands

from udo.flags import Flags


async def send_traceback(
    destination: typing.Union[discord.abc.Messageable, discord.Message],
    verbosity: int,
    etype: typing.Type[BaseException],
    value: BaseException,
    trace: TracebackType
):
    """
    Sends a traceback of an exception to a destination.
    Used when REPL fails for any reason.
    :param destination: Where to send this information to
    :param verbosity: How far back this traceback should go. 0 shows just the last stack.
    :param exc_info: Information about this exception, from sys.exc_info or similar.
    :return: The last message sent
    """

    traceback_content = "".join(traceback.format_exception(etype, value, trace, verbosity)).replace("``", "`\u200b`")

    paginator = commands.Paginator(prefix='```py')
    for line in traceback_content.split('\n'):
        paginator.add_line(line)

    message = None

    for page in paginator.pages:
        if isinstance(destination, discord.Message):
            message = await destination.reply(page)
        else:
            message = await destination.send(page)

    return message


T = typing.TypeVar('T')

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
    P = ParamSpec('P')
else:
    P = typing.ParamSpec('P')  # pylint: disable=no-member


async def do_after_sleep(delay: float, coro: typing.Callable[P, typing.Awaitable[T]], *args: P.args, **kwargs: P.kwargs) -> T:
    """
    Performs an action after a set amount of time.
    This function only calls the coroutine after the delay,
    preventing asyncio complaints about destroyed coros.
    :param delay: Time in seconds
    :param coro: Coroutine to run
    :param args: Arguments to pass to coroutine
    :param kwargs: Keyword arguments to pass to coroutine
    :return: Whatever the coroutine returned.
    """
    await asyncio.sleep(delay)
    return await coro(*args, **kwargs)


async def attempt_add_reaction(
    msg: discord.Message,
    reaction: typing.Union[str, discord.Emoji]
) -> typing.Optional[discord.Reaction]:
    """
    Try to add a reaction to a message, ignoring it if it fails for any reason.
    :param msg: The message to add the reaction to.
    :param reaction: The reaction emoji, could be a string or `discord.Emoji`
    :return: A `discord.Reaction` or None, depending on if it failed or not.
    """
    try:
        return await msg.add_reaction(reaction)
    except discord.HTTPException:
        pass


class ReplResponseReactor:  # pylint: disable=too-few-public-methods
    """
    Extension of the ReactionProcedureTimer that absorbs errors, sending tracebacks.
    """

    __slots__ = ('message', 'loop', 'handle', 'raised')

    def __init__(self, message: discord.Message, loop: typing.Optional[asyncio.BaseEventLoop] = None):
        self.message = message
        self.loop = loop or asyncio.get_event_loop()
        self.handle = None
        self.raised = False

    async def __aenter__(self):
        self.handle = self.loop.create_task(do_after_sleep(2, attempt_add_reaction, self.message,
                                                           "\N{BLACK RIGHT-POINTING TRIANGLE}"))
        return self

    async def __aexit__(
        self,
        exc_type: typing.Type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType
    ) -> bool:
        if self.handle:
            self.handle.cancel()

        # no exception, check mark
        if not exc_val:
            await attempt_add_reaction(self.message, "\N{WHITE HEAVY CHECK MARK}")
            return False

        self.raised = True

        if isinstance(exc_val, (SyntaxError, asyncio.TimeoutError, subprocess.TimeoutExpired)):
            # short traceback, send to channel
            destination = Flags.traceback_destination(self.message) or self.message.channel

            if destination != self.message.channel:
                await attempt_add_reaction(
                    self.message,
                    # timed out is alarm clock
                    # syntax error is single exclamation mark
                    "\N{HEAVY EXCLAMATION MARK SYMBOL}" if isinstance(exc_val, SyntaxError) else "\N{ALARM CLOCK}"
                )

            await send_traceback(
                self.message if destination == self.message.channel else destination,
                0, exc_type, exc_val, exc_tb
            )
        else:
            destination = Flags.traceback_destination(self.message) or self.message.author

            if destination != self.message.channel:
                # other error, double exclamation mark
                await attempt_add_reaction(self.message, "\N{DOUBLE EXCLAMATION MARK}")

            # this traceback likely needs more info, so increase verbosity, and DM it instead.
            await send_traceback(
                self.message if destination == self.message.channel else destination,
                8, exc_type, exc_val, exc_tb
            )

        return True  # the exception has been handled


import discord 
from discord.ext import commands
import sys
import typing
import importlib.metadata
from .work.types import ContextA
from .codeblocks import Codeblock, codeblock_converter
import traceback
import math
import itertools
import time
from urllib.parse import urlencode
import asyncio
import io
import os, json
from datetime import datetime, timezone
import logging


from udo.modules import ExtensionConverter
from udo.shell import ShellReader
from udo.exception_handling import ReplResponseReactor
from udo.flags import Flags
from udo.paginators import PaginatorInterface, WrappedPaginator, use_file_check
from udo.functools import AsyncSender
from udo.repl import AsyncCodeExecutor, Scope, all_inspections, disassemble, get_var_dict_from_ctx

Directory = os.path.dirname(os.path.realpath(__file__))
logger: logging.Logger = logging.getLogger(__name__)

# ==============================================================================

try:
    import psutil
except ImportError:
    psutil = None

try:
    from importlib.metadata import distribution, packages_distributions
except ImportError:
    from importlib_metadata import distribution, packages_distributions

# ==============================================================================





def natural_size(size_in_bytes: int) -> str:
    """
    Converts a number of bytes to an appropriately-scaled unit
    E.g.:
        1024 -> 1.00 KiB
        12345678 -> 11.77 MiB
    """
    units = ('B', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB', 'YiB')

    power = int(math.log(max(abs(size_in_bytes), 1), 1024))

    return f"{size_in_bytes / (1024 ** power):.2f} {units[power]}"


def natural_time(time_in_seconds: float) -> str:
    """
    Converts a time in seconds to a 6-padded scaled unit
    E.g.:
        1.5000 ->   1.50  s
        0.1000 -> 100.00 ms
        0.0001 -> 100.00 us
    """
    units = (
        ('mi', 60),
        (' s', 1),
        ('ms', 1e-3),
        ('\N{GREEK SMALL LETTER MU}s', 1e-6),
    )

    absolute = abs(time_in_seconds)

    for label, size in units:
        if absolute > size:
            return f"{time_in_seconds / size:6.2f} {label}"

    return f"{time_in_seconds / 1e-9:6.2f} ns"


def mean_stddev(collection: typing.Collection[float]) -> typing.Tuple[float, float]:
    """
    Takes a collection of floats and returns (mean, stddev) as a tuple.
    """

    average = sum(collection) / len(collection)

    if len(collection) > 1:
        stddev = math.sqrt(sum(math.pow(reading - average, 2) for reading in collection) / (len(collection) - 1))
    else:
        stddev = 0.0

    return (average, stddev)


def format_stddev(collection: typing.Collection[float]) -> str:
    """
    Takes a collection of floats and produces a mean (+ stddev, if multiple values exist) string.
    """
    if len(collection) > 1:
        average, stddev = mean_stddev(collection)

        return f"{natural_time(average)} \N{PLUS-MINUS SIGN} {natural_time(stddev)}"

    return natural_time(sum(collection) / len(collection))


class Dropdown(discord.ui.Select):
    def __init__(self, bot_: commands.Bot, requester,start_time):
        self.bot = bot_
        self.requester = requester
        self.start_time = start_time

        options = [
            discord.SelectOption(
                label="udo page", description="View the Jeju Island main page.", emoji="🟥"
            ),
            discord.SelectOption(
                label="protected_access page", description="The page related to the intent.", emoji="🟩"
            )
        ]

        super().__init__(
            placeholder="udo !! Click here to open the menu!",
            min_values=1,
            max_values=1,
            options=options,
        )

    async def callback(self, interaction: discord.Interaction):
        if interaction.user != self.requester:
            return await interaction.response.send_message("❌ udo is an owner-only command. !!", ephemeral=True)

        if self.values[0] == "udo page":

            udo = [
                f"> the, udo ` v{importlib.metadata.version('udo.py')} `,  discord.py ` v{importlib.metadata.version('discord.py')} `".replace("\n", ""),
                f"ㄴ • Python ` {sys.version}`".replace("\n", ""),
                f"ㄴ • platform `{sys.platform}` • start loaded <t:{self.start_time.timestamp():.0f}:R>".replace("\n", "")
            ]
            
            if psutil:
                try:
                    proc = psutil.Process()

                    with proc.oneshot():
                        try:
                            mem = proc.memory_full_info()
                            udo.append(f"\n • {natural_size(mem.rss)} memory , {natural_size(mem.vms)} memory , {natural_size(mem.uss)} memory")
                        except psutil.AccessDenied:
                            pass

                        try:
                            name = proc.name()
                            pid = proc.pid
                            thread_count = proc.num_threads()

                            udo.append(f"ㄴ • PID {pid} (`{name}`) == {thread_count} thread(s)")
                        except psutil.AccessDenied:
                            pass

                        udo.append("")
                except psutil.AccessDenied:
                    udo.append(
                        "ㄴ • psutil is installed, but this process does not have high enough access rights "
                    )
                    udo.append("")

            cache_summary = f"` 🌐 {len(self.bot.guilds)} ` guild(s) ㅣ ` 🤴 {len(self.bot.users)} ` user(s)"

            if isinstance(self.bot, discord.AutoShardedClient):
                if len(self.bot.shards) > 20:
                    udo.append(
                        f"This bot is auto shards (` {len(self.bot.shards)} ` shards of ` {self.bot.shard_count} `)"
                        f"\n{cache_summary}."
                    )
                else:
                    shard_ids = ', '.join(str(i) for i in self.bot.shards.keys())
                    udo.append(
                        f"This bot is auto shards (Shards ` {shard_ids} ` of ` {self.bot.shard_count} `)"
                        f"\n{cache_summary}."
                    )
            elif self.bot.shard_count:
                udo.append(
                    f"This bot is manually sharded (Shard ` {self.bot.shard_id} ` of ` {self.bot.shard_count} `)"
                    f"\n{cache_summary}."
                )
            else:
                udo.append(f"This bot is not sharded\n{cache_summary}.")

            udo.append(f"\n\nAverage websocket latency: {round(self.bot.latency * 1000, 2)}ms")

            await interaction.response.edit_message(
                content="\n".join(udo)
            )
        if self.values[0] == "protected_access page":
            udo = [
                f"> the, udo ` v{importlib.metadata.version('udo.py')} `,  discord.py ` v{importlib.metadata.version('discord.py')} `".replace("\n", ""),
                f"ㄴ • Python ` {sys.version}`".replace("\n", ""),
                f"ㄴ • platform `{sys.platform}` • start loaded <t:{self.start_time.timestamp():.0f}:R>".replace("\n", "")
            ]

            if self.bot._connection.max_messages:
                message_cache = f"Message cache capped at {self.bot._connection.max_messages}"
            else:
                message_cache = "Message cache is disabled"

            if discord.version_info >= (1, 5, 0):
                remarks = {
                    True: 'enabled',
                    False: 'disabled',
                    None: 'unknown'
                }

                *group, last = (
                    f"{intent.replace('_', ' ')} intent is {remarks.get(getattr(self.bot.intents, intent, None))}\n"
                    for intent in
                    ('presences', 'members', 'message_content')
                )

                udo.append(f"\n{message_cache}\n{''.join(group)}\n{last}")
            else:
                guild_subscriptions = f"guild subscriptions are {'enabled' if self.bot._connection.guild_subscriptions else 'disabled'}"  # type: ignore

                udo.append(f"\n{message_cache}\n{guild_subscriptions}")
            udo.append(f"Average websocket latency: {round(self.bot.latency * 1000, 2)}ms")

            await interaction.response.edit_message(
                content="\n".join(udo)
            )

class DropdownView(discord.ui.View):
    def __init__(self, bot_: commands.Bot, requester, start_time):
        self.bot = bot_
        self.requester = requester
        self.start_time = start_time
        super().__init__()
        self.add_item(Dropdown(self.bot, self.requester, self.start_time))



class udo(commands.Cog):
    def __init__(self, bot):
        self.bot: commands.Bot = bot
        self._scope = Scope()
        self.retain = Flags.RETAIN
        self.last_result = None
        self.start_time: datetime = datetime.utcnow().replace(tzinfo=timezone.utc)
        

    @property
    def scope(self):
        """
        Gets a scope for use in REPL.
        If retention is on, this is the internal stored scope,
        otherwise it is always a new Scope.
        """

        if self.retain:
            return self._scope
        return Scope()

    @commands.Cog.listener()
    async def on_command_error(self, ctx, error):
        if isinstance(error, commands.NotOwner):
            await ctx.reply(f"❌ ` {ctx.command.name} ` is an owner-only command. !!")


    async def jsk_python_result_handling(self, ctx: commands.Context, result):  # pylint: disable=too-many-return-statements
        """
        Determines what is done with a result when it comes out of jsk py.
        This allows you to override how this is done without having to rewrite the command itself.
        What you return is what gets stored in the temporary _ variable.
        """

        if isinstance(result, discord.Message):
            return await ctx.send(f"<Message <{result.jump_url}>>")

        if isinstance(result, discord.File):
            return await ctx.send(file=result)

        if isinstance(result, discord.Embed):
            return await ctx.send(embed=result)

        if isinstance(result, PaginatorInterface):
            return await result.send_to(ctx)

        if not isinstance(result, str):
            # repr all non-strings
            result = repr(result)

        # Eventually the below handling should probably be put somewhere else
        if len(result) <= 2000:
            if result.strip() == "":
                result = "\u200b"

            return await ctx.send(result.replace(self.bot.http.token, "[token omitted]"))

        if use_file_check(ctx, len(result)):  # File "full content" preview limit
            # Discord's desktop and web client now supports an interactive file content
            #  display for files encoded in UTF-8.
            # Since this avoids escape issues and is more intuitive than pagination for
            #  long results, it will now be prioritized over PaginatorInterface if the
            #  resultant content is below the filesize threshold
            return await ctx.send(file=discord.File(filename="output.py", fp=io.BytesIO(result.encode("utf-8"))))

        # inconsistency here, results get wrapped in codeblocks when they are too large
        #  but don't if they're not. probably not that bad, but noting for later review
        paginator = WrappedPaginator(prefix="```py", suffix="```", max_size=1985)

        paginator.add_line(result)

        interface = PaginatorInterface(ctx.bot, paginator, owner=ctx.author)
        return await interface.send_to(ctx)









# udo

    @commands.group("udo",aliases=["우도"],invoke_without_command=True)
    @commands.is_owner()
    async def udo_jsk(self, ctx: ContextA):
        """
        The udo debug and diagnostic commands.
        This command on its own gives a status brief.
        All other functionality is within its subcommands.
        """
        try:

            udo = [
                f"> the, udo ` v{importlib.metadata.version('udo.py')} `,  discord.py ` v{importlib.metadata.version('discord.py')} `".replace("\n", ""),
                f"ㄴ • Python ` {sys.version}`".replace("\n", ""),
                f"ㄴ • platform `{sys.platform}` • start loaded <t:{self.start_time.timestamp():.0f}:R>".replace("\n", "")
            ]
            
            if psutil:
                try:
                    proc = psutil.Process()

                    with proc.oneshot():
                        try:
                            mem = proc.memory_full_info()
                            udo.append(f"\n • {natural_size(mem.rss)} memory , {natural_size(mem.vms)} memory , {natural_size(mem.uss)} memory")
                        except psutil.AccessDenied:
                            pass

                        try:
                            name = proc.name()
                            pid = proc.pid
                            thread_count = proc.num_threads()

                            udo.append(f"ㄴ • PID {pid} (`{name}`) == {thread_count} thread(s)")
                        except psutil.AccessDenied:
                            pass

                        udo.append("")  # blank line
                except psutil.AccessDenied:
                    udo.append(
                        "ㄴ • psutil is installed, but this process does not have high enough access rights "
                    )
                    udo.append("")  # blank line

            cache_summary = f"` 🌐 {len(self.bot.guilds)} ` guild(s) ㅣ ` 🤴 {len(self.bot.users)} ` user(s)"

            # Show shard settings to summary
            if isinstance(self.bot, discord.AutoShardedClient):
                if len(self.bot.shards) > 20:
                    udo.append(
                        f"This bot is auto shards (` {len(self.bot.shards)} ` shards of ` {self.bot.shard_count} `)"
                        f"\n{cache_summary}."
                    )
                else:
                    shard_ids = ', '.join(str(i) for i in self.bot.shards.keys())
                    udo.append(
                        f"This bot is auto shards (Shards ` {shard_ids} ` of ` {self.bot.shard_count} `)"
                        f"\n{cache_summary}."
                    )
            elif self.bot.shard_count:
                udo.append(
                    f"This bot is manually sharded (Shard ` {self.bot.shard_id} ` of ` {self.bot.shard_count} `)"
                    f"\n{cache_summary}."
                )
            else:
                udo.append(f"This bot is not sharded\n{cache_summary}.")
            udo.append(f"\n\nAverage websocket latency: {round(self.bot.latency * 1000, 2)}ms")

            view = DropdownView(self.bot, ctx.author,self.start_time)
            await ctx.reply("\n".join(udo), view=view)
        except:
            logger.error(traceback.format_exc())

# udo shell

    @udo_jsk.command(name="shell", aliases=["bash", "sh", "powershell", "ps1", "ps", "cmd", "terminal"])
    @commands.is_owner()
    async def udo_shell(self, ctx: ContextA, *, argument: codeblock_converter=None):
        """
        Run a shell command
        """
        try:
            if not argument:
                await ctx.send("Usage: `<prefix> udo shell <argument>`")
                return
            if typing.TYPE_CHECKING:
                argument: Codeblock = argument  # type: ignore

            try:
                async with ReplResponseReactor(ctx.message):
                    with ShellReader(argument.content) as reader:
                        prefix = "```" + reader.highlight

                        paginator = WrappedPaginator(prefix=prefix, max_size=1975)
                        paginator.add_line(f"{reader.ps1} {argument.content}\n")

                        interface = PaginatorInterface(ctx.bot, paginator, owner=ctx.author)
                        self.bot.loop.create_task(interface.send_to(ctx))

                        async for line in reader:
                            if interface.closed:
                                return
                            await interface.add_line(line)

                    await interface.add_line(f"\n[ jeju status ] Return code {reader.close_code}")
            except:
                logger.error(traceback.format_exc())
        except asyncio.TimeoutError:
            pass

# udo shutdown

    @udo_jsk.command(name="shutdown", aliases=["logout", "out", "doun"])
    @commands.is_owner()
    async def udo_shutdown(self, ctx: ContextA):
        """
        Logs this bot out.
        """
        try:

            await ctx.send(f"__**[ ` udo ` ]** •  Shutdown now . . .__")
            await ctx.bot.close()
        except:
            logger.error(traceback.format_exc())

# udo rtt

    @udo_jsk.command(name="rtt", aliases=["ping"])
    @commands.is_owner()
    async def udo_rtt(self, ctx: ContextA):
        """
        Calculates Round-Trip Time to the API.
        """
        try:
            message = None

            # We'll show each of these readings as well as an average and standard deviation.
            api_readings: typing.List[float] = []
            # We'll also record websocket readings, but we'll only provide the average.
            websocket_readings: typing.List[float] = []

            # We do 6 iterations here.
            # This gives us 5 visible readings, because a request can't include the stats for itself.
            for _ in range(6):
                # First generate the text
                text = ""
                text += "\n".join(f"Reading {index + 1}: {reading * 1000:.2f}ms" for index, reading in enumerate(api_readings))

                if api_readings:
                    average, stddev = mean_stddev(api_readings)

                    text += f"\n\nAverage: {average * 1000:.2f} \N{PLUS-MINUS SIGN} {stddev * 1000:.2f}ms"
                else:
                    text += "\n\nNo readings yet."

                if websocket_readings:
                    average = sum(websocket_readings) / len(websocket_readings)

                    text += f"\nWebsocket latency: {average * 1000:.2f}ms"
                else:
                    text += f"\nWebsocket latency: {self.bot.latency * 1000:.2f}ms"

                # Now do the actual request and reading
                if message:
                    before = time.perf_counter()
                    await message.edit(content=text)
                    after = time.perf_counter()

                    api_readings.append(after - before)
                else:
                    before = time.perf_counter()
                    message = await ctx.send(content=text)
                    after = time.perf_counter()

                    api_readings.append(after - before)

                # Ignore websocket latencies that are 0 or negative because they usually mean we've got bad heartbeats
                if self.bot.latency > 0.0:
                    websocket_readings.append(self.bot.latency)
        except:
            logger.error(traceback.format_exc())

# udo load

    @udo_jsk.command(name="load", aliases=["reload"])
    @commands.is_owner()
    async def udo_load(self, ctx: ContextA, *extensions: ExtensionConverter):  # type: ignore
        # """
        # Loads or reloads the given extension names.
        # Reports any extensions that failed to load.
        # """
        # try:
        #     extensions: typing.Iterable[typing.List[str]] = extensions  # type: ignore

        #     paginator = commands.Paginator(prefix='', suffix='')

        #     # 'jsk reload' on its own just reloads jishaku
        #     if ctx.invoked_with == 'reload' and not extensions:
        #         extensions = [['udo']]

        #     for extension in itertools.chain(*extensions):
        #         method, icon = (
        #             (self.bot.reload_extension, "\N{CLOCKWISE RIGHTWARDS AND LEFTWARDS OPEN CIRCLE ARROWS}")
        #             if extension in self.bot.extensions else
        #             (self.bot.load_extension, "\N{INBOX TRAY}")
        #         )

        #         try:
        #             await discord.utils.maybe_coroutine(method, extension)
        #         except Exception as exc:  # pylint: disable=broad-except
        #             if isinstance(exc, commands.ExtensionFailed) and exc.__cause__:
        #                 cause = exc.__cause__
        #                 traceback_data = ''.join(traceback.format_exception(type(cause), cause, cause.__traceback__, 8))
        #             else:
        #                 traceback_data = ''.join(traceback.format_exception(type(exc), exc, exc.__traceback__, 2))

        #             paginator.add_line(
        #                 f"{icon}\N{WARNING SIGN} `{extension}`\n```py\n{traceback_data}\n```",
        #                 empty=True
        #             )
        #         else:
        #             paginator.add_line(f"{icon} `{extension}`", empty=True)

        #     for page in paginator.pages:
        #         await ctx.send(page)
        # except:
        #     logger.error(traceback.format_exc())

        """
        Loads or reloads the given extension names.

        Reports any extensions that failed to load.
        """
        try:
            paginator = WrappedPaginator(prefix="", suffix="")

            # 'jsk reload' on its own just reloads jishaku
            if ctx.invoked_with == "reload" and not extensions:
                extensions = [["udo"]]

            for extension in itertools.chain(*extensions):
                method, icon = (
                    (
                        self.bot.reload_extension,
                        "\N{CLOCKWISE RIGHTWARDS AND LEFTWARDS OPEN CIRCLE ARROWS}",
                    )
                    if extension in self.bot.extensions
                    else (self.bot.load_extension, "\N{INBOX TRAY}")
                )

                try:
                    method(extension)
                except Exception as exc:  # pylint: disable=broad-except
                    traceback_data = "".join(traceback.format_exception(type(exc), exc, exc.__traceback__, 1))

                    paginator.add_line(
                        f"{icon}\N{WARNING SIGN} `{extension}`\n```py\n{traceback_data}\n```",
                        empty=True,
                    )
                else:
                    paginator.add_line(f"{icon} `{extension}`", empty=True)

            for page in paginator.pages:
                await ctx.send(page)
        except:
            logger.error(traceback.format_exc())

# udo unload

    @udo_jsk.command(name="unload")
    @commands.is_owner()
    async def udo_unload(self, ctx: commands.Context, *extensions: ExtensionConverter):
        """
        Unloads the given extension names.
        Reports any extensions that failed to unload.
        """
        try:
            paginator = WrappedPaginator(prefix="", suffix="")
            icon = "\N{OUTBOX TRAY}"

            for extension in itertools.chain(*extensions):
                try:
                    self.bot.unload_extension(extension)
                except Exception as exc:  # pylint: disable=broad-except
                    traceback_data = "".join(traceback.format_exception(type(exc), exc, exc.__traceback__, 1))

                    paginator.add_line(
                        f"{icon}\N{WARNING SIGN} `{extension}`\n```py\n{traceback_data}\n```",
                        empty=True,
                    )
                else:
                    paginator.add_line(f"{icon} `{extension}`", empty=True)

            for page in paginator.pages:
                await ctx.send(page)
        except:
            logger.error(traceback.format_exc())

# udo python

    @udo_jsk.command(name="python", aliases=["py", "eval"])
    @commands.is_owner()
    async def udo_python(self, ctx: ContextA, *, argument: codeblock_converter):
        """
        Direct evaluation of Python code.
        """
        try:
            arg_dict = get_var_dict_from_ctx(ctx, Flags.SCOPE_PREFIX)
            arg_dict["_"] = self.last_result

            scope = self.scope

            try:
                async with ReplResponseReactor(ctx.message):

                    executor = AsyncCodeExecutor(argument.content, scope, arg_dict=arg_dict)
                    async for send, result in AsyncSender(executor):
                        if result is None:
                            continue

                        self.last_result = result

                        send(await self.jsk_python_result_handling(ctx, result))

            finally:
                scope.clear_intersection(arg_dict)
        except:
            logger.error(traceback.format_exc())

# tag

    @commands.group("tag", invoke_without_command=True)
    async def tag_tag(self, ctx, tag=None):
        try:
            if not tag:
                await ctx.send("Usage: `<prefix> tag <tag>`")
                return
            
            with open(f"{Directory}/data.json", "r",encoding="utf-8-sig") as json_file:
                dbb=json.load(json_file)
            if tag in dbb.keys():
                return await ctx.send(f"{dbb[str(tag)]['value']}")
            else:
                return await ctx.send(f"The tag {tag} could not be found.")
                
        except:
            logger.error(traceback.format_exc())
            
# tag add
    
    @tag_tag.command(name="add")
    @commands.is_owner()
    async def tag_add(self, ctx, tag_name, *, tag_add):
        try:
            try:
                with open(f"{Directory}/data.json", "r",encoding="utf-8-sig") as json_file:
                    dbb=json.load(json_file)
                if dbb[str(tag_name)]["tag"] == tag_name:
                    return await ctx.send(f"`❌` Sorry.. This tag is already fixed.")
            except:
                with open(f"{Directory}/data.json", "r",encoding="utf-8-sig") as json_file:
                    dbb=json.load(json_file)
                dbb[str(tag_name)] ={
                    "tag" : str(tag_name),
                    "value" : str(tag_add)
                }
                with open(f"{Directory}/data.json", "w",encoding="utf-8-sig") as json_file:
                    json.dump(dbb,json_file,ensure_ascii = False, indent=4)
                return await ctx.send(f"`✅` Successfully created tag\n> **Tag_Name** : {tag_name}\n> **Tag_Value** : {tag_add}")
        except:
            logger.error(traceback.format_exc())

# tag update

    @tag_tag.command(name="update", aliases=["uptag"])
    @commands.is_owner()
    async def tag_update(self, ctx, tag_name, *, tag_add):
        try:
            with open(f"{Directory}/data.json", "r",encoding="utf-8-sig") as json_file:
                dbb=json.load(json_file)
            if tag_name in dbb.keys():
                dbb[str(tag_name)]["value"] = str(tag_add)
                with open(f"{Directory}/data.json", "w",encoding="utf-8-sig") as json_file:
                    json.dump(dbb,json_file,ensure_ascii = False, indent=4)
                return await ctx.send(f"`✅` Successfully update tag\n> **Tag_Name** : {tag_name}\n> **Tag_Value** : {tag_add}")
            else:
                return await ctx.send(f"The tag {tag_name} could not be found.")
        
        except:
            logger.error(traceback.format_exc())

# tag list

    @tag_tag.command(name="list")
    # @commands.is_owner()
    async def tag_list(self, ctx):
        try:
            with open(f"{Directory}/data.json", "r",encoding="utf-8-sig") as json_file:
                dbb=json.load(json_file)
            lists = list()
            for i in dbb:
                lists.append(f'{i}')
            lists = ", ".join(lists)
            return await ctx.send(f"```\n{lists}\n```")
        except:
            logger.error(traceback.format_exc())


import typing

import discord
from discord.ext import commands

BotT = typing.Union[commands.Bot, commands.AutoShardedBot]

if typing.TYPE_CHECKING or discord.version_info >= (2, 0, 0):
    ContextT = typing.TypeVar('ContextT', commands.Context[commands.Bot], commands.Context[commands.AutoShardedBot])
    ContextA = commands.Context[BotT]
else:
    ContextT = ContextA = commands.Context


# -*- coding: utf-8 -*-

"""
jishaku.repl.disassembly
~~~~~~~~~~~~~~~~~~~~~~~~

Functions pertaining to the disassembly of Python code

:copyright: (c) 2021 Devon (Gorialis) R
:license: MIT, see LICENSE for more details.

"""

import ast
import dis

import import_expression

from udo.repl.scope import Scope

CORO_CODE = f"""
import asyncio

import discord
from discord.ext import commands
from importlib import import_module as {import_expression.constants.IMPORTER}

import jishaku

async def _repl_coroutine({{0}}):
    pass
"""


def wrap_code(code: str, args: str = "") -> ast.Module:
    """
    Wraps code for disassembly.

    This is similar in function to the jishaku.repl.compilation equivalent,
    but due to the different structure required for clean disassemblies,
    it's implemented separately here.
    """

    user_code = import_expression.parse(code, mode="exec")
    mod = import_expression.parse(CORO_CODE.format(args), mode="exec")

    definition = mod.body[-1]  # async def ...:
    assert isinstance(definition, ast.AsyncFunctionDef)

    # Patch user code directly into the function
    definition.body = user_code.body

    ast.fix_missing_locations(mod)

    # We do not use the keyword transformer here, since it might produce misleading disassembly.

    is_asyncgen = any(isinstance(node, ast.Yield) for node in ast.walk(definition))
    last_expr = definition.body[-1]

    # if the last part isn't an expression, ignore it
    if not isinstance(last_expr, ast.Expr):
        return mod

    # if this isn't a generator and the last expression is not a return
    if not is_asyncgen and not isinstance(last_expr.value, ast.Return):
        # copy the value of the expression into a return
        return_stmt = ast.Return(last_expr.value)
        ast.copy_location(return_stmt, last_expr)

        # place the return where the original expression was
        definition.body[-1] = return_stmt

    return mod


def disassemble(code: str, scope: Scope = None, arg_dict: dict = None):
    """
    Disassembles asynchronous code into dis.dis-style bytecode instructions.
    """

    # Similar to AsyncCodeExecutor.__init__
    arg_names = list(arg_dict.keys()) if arg_dict else []

    scope = scope or Scope()

    wrapped = wrap_code(code, args=", ".join(arg_names))
    exec(compile(wrapped, "<repl>", "exec"), scope.globals, scope.locals)  # pylint: disable=exec-used

    func_def = scope.locals.get("_repl_coroutine") or scope.globals["_repl_coroutine"]

    # pylint is gonna really hate this part onwards
    # pylint: disable=protected-access, invalid-name
    co = func_def.__code__

    for instruction in dis._get_instructions_bytes(
        co.co_code,
        co.co_varnames,
        co.co_names,
        co.co_consts,
        co.co_cellvars + co.co_freevars,
        dict(dis.findlinestarts(co)),
        line_offset=0,
    ):
        if instruction.starts_line is not None and instruction.offset > 0:
            yield ""

        yield instruction._disassemble(4, False, 4)

    # pylint: enable=protected-access, invalid-name




import typing

import discord
from discord.ext import commands

from udo.flags import Flags
from udo.hljs import get_language, guess_file_traits
from udo.core.shim.paginator_base import EmojiSettings
from udo.core.work.types import ContextA

# Version detection
if discord.version_info >= (2, 0, 0):
    from udo.core.shim.paginator_200 import PaginatorEmbedInterface, PaginatorInterface
else:
    from udo.core.shim.paginator_170 import PaginatorEmbedInterface, PaginatorInterface

__all__ = ('EmojiSettings', 'PaginatorInterface', 'PaginatorEmbedInterface',
           'WrappedPaginator', 'FilePaginator', 'use_file_check')


class WrappedPaginator(commands.Paginator):
    """
    A paginator that allows automatic wrapping of lines should they not fit.
    This is useful when paginating unpredictable output,
    as it allows for line splitting on big chunks of data.
    Delimiters are prioritized in the order of their tuple.
    Parameters
    -----------
    wrap_on: tuple
        A tuple of wrapping delimiters.
    include_wrapped: bool
        Whether to include the delimiter at the end of a wrapped line.
    force_wrap: bool
        If this is True, lines will be split at their maximum points should trimming not be possible
        with any provided delimiter.
    """

    def __init__(
        self,
        *args: typing.Any,
        wrap_on: typing.Tuple[str, ...] = ('\n', ' '),
        include_wrapped: bool = True,
        force_wrap: bool = False,
        **kwargs: typing.Any
    ):
        super().__init__(*args, **kwargs)
        self.wrap_on = wrap_on
        self.include_wrapped = include_wrapped
        self.force_wrap = force_wrap

    def add_line(self, line: str = '', *, empty: bool = False):
        true_max_size = self.max_size - self._prefix_len - self._suffix_len - 2 * self._linesep_len
        start = 0
        needle = 0
        last_delimiter = -1
        last_space = -1

        while needle < len(line):
            if needle - start >= true_max_size:
                if last_delimiter != -1:
                    if self.include_wrapped and line[last_delimiter] != '\n':
                        super().add_line(line[start:last_delimiter + 1])
                        needle = last_delimiter + 1
                        start = last_delimiter + 1
                    else:
                        super().add_line(line[start:last_delimiter])
                        needle = last_delimiter + 1
                        start = last_delimiter + 1
                elif last_space != -1:
                    super().add_line(line[start:last_space])
                    needle = last_space + 1
                    start = last_space
                else:
                    super().add_line(line[start:needle])
                    start = needle

                last_delimiter = -1
                last_space = -1

            if line[needle] in self.wrap_on:
                last_delimiter = needle
            elif line[needle] == ' ':
                last_space = needle

            needle += 1

        last_line = line[start:needle]
        if last_line:
            super().add_line(last_line)

        if empty:
            self._current_page.append('')
            self._count += self._linesep_len


class FilePaginator(commands.Paginator):
    """
    A paginator of syntax-highlighted codeblocks, read from a file-like.
    Parameters
    -----------
    fp
        A file-like (implements ``fp.read``) to read the data for this paginator from.
    line_span: Optional[Tuple[int, int]]
        A linespan to read from the file. If None, reads the whole file.
    language_hints: Tuple[str, ...]
        A tuple of strings that may hint to the language of this file.
        This could include filenames, MIME types, or shebangs.
        A shebang present in the actual file will always be prioritized over this.
    """

    def __init__(
        self,
        fp: typing.BinaryIO,
        line_span: typing.Optional[typing.Tuple[int, int]] = None,
        language_hints: typing.Tuple[str, ...] = (),
        **kwargs: typing.Any
    ):
        language = ''

        for hint in language_hints:
            language = get_language(hint)

            if language:
                break

        if not language:
            try:
                language = get_language(fp.name)
            except AttributeError:
                pass

        content, _, file_language = guess_file_traits(fp.read())

        language = file_language or language
        lines = content.split('\n')

        super().__init__(prefix=f'```{language}', suffix='```', **kwargs)

        if line_span:
            if line_span[1] < line_span[0]:
                line_span = (line_span[1], line_span[0])

            if line_span[0] < 1 or line_span[1] > len(lines):
                raise ValueError("Linespan goes out of bounds.")

            lines = lines[line_span[0] - 1:line_span[1]]

        for line in lines:
            self.add_line(line)


class WrappedFilePaginator(FilePaginator, WrappedPaginator):
    """
    Combination of FilePaginator and WrappedPaginator.
    In other words, a FilePaginator that supports line wrapping.
    """


def use_file_check(
    ctx: ContextA,
    size: int
) -> bool:
    """
    A check to determine if uploading a file and relying on Discord's file preview is acceptable over a PaginatorInterface.
    """

    return all([
        size < 50_000,  # Check the text is below the Discord cutoff point;
        not Flags.FORCE_PAGINATOR,  # Check the user hasn't explicitly disabled this;
        (
            # Ensure the user isn't on mobile
            not ctx.author.is_on_mobile()
            if ctx.guild and ctx.bot.intents.presences and isinstance(ctx.author, discord.Member)
            else True
        )
    ])


# -*- coding: utf-8 -*-



# pylint: disable=wildcard-import
from udo.repl.compilation import *  # noqa: F401
from udo.repl.disassembly import disassemble  # noqa: F401
from udo.repl.inspections import all_inspections  # noqa: F401
from udo.repl.repl_builtins import get_var_dict_from_ctx  # noqa: F401
from udo.repl.scope import *  # noqa: F401



import asyncio
import os
import pathlib
import re
import subprocess
import sys
import time
import typing

T = typing.TypeVar('T')

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
    P = ParamSpec('P')
else:
    P = typing.ParamSpec('P')  # pylint: disable=no-member


SHELL = os.getenv("SHELL") or "/bin/bash"
WINDOWS = sys.platform == "win32"


def background_reader(stream: typing.IO[bytes], loop: asyncio.AbstractEventLoop, callback: typing.Callable[[bytes], typing.Any]):
    """
    Reads a stream and forwards each line to an async callback.
    """

    for line in iter(stream.readline, b''):
        loop.call_soon_threadsafe(loop.create_task, callback(line))


class ShellReader:
    """
    A class that passively reads from a shell and buffers results for read.
    Example
    -------
    .. code:: python3
        # reader should be in a with statement to ensure it is properly closed
        with ShellReader('echo one; sleep 5; echo two') as reader:
            # prints 'one', then 'two' after 5 seconds
            async for x in reader:
                print(x)
    """

    def __init__(
        self,
        code: str,
        timeout: int = 120,
        loop: typing.Optional[asyncio.AbstractEventLoop] = None,
        escape_ansi: bool = True
    ):
        if WINDOWS:
            # Check for powershell
            if pathlib.Path(r"C:\Windows\System32\WindowsPowerShell\v1.0\powershell.exe").exists():
                sequence = ['powershell', code]
                self.ps1 = "PS >"
                self.highlight = "powershell"
            else:
                sequence = ['cmd', '/c', code]
                self.ps1 = "cmd >"
                self.highlight = "cmd"
            # Windows doesn't use ANSI codes
            self.escape_ansi = True
        else:
            sequence = [SHELL, '-c', code]
            self.ps1 = "$"
            self.highlight = "ansi"
            self.escape_ansi = escape_ansi

        self.process = subprocess.Popen(sequence, stdout=subprocess.PIPE, stderr=subprocess.PIPE)  # pylint: disable=consider-using-with
        self.close_code = None

        self.loop = loop or asyncio.get_event_loop()
        self.timeout = timeout

        self.stdout_task = self.make_reader_task(self.process.stdout, self.stdout_handler) if self.process.stdout else None
        self.stderr_task = self.make_reader_task(self.process.stderr, self.stderr_handler) if self.process.stderr else None

        self.queue: asyncio.Queue[str] = asyncio.Queue(maxsize=250)

    @property
    def closed(self) -> bool:
        """
        Are both tasks done, indicating there is no more to read?
        """

        return (not self.stdout_task or self.stdout_task.done()) and (not self.stderr_task or self.stderr_task.done())

    async def executor_wrapper(self, func: typing.Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
        """
        Call wrapper for stream reader.
        """

        return await self.loop.run_in_executor(None, func, *args, **kwargs)

    def make_reader_task(self, stream: typing.IO[bytes], callback: typing.Callable[[bytes], typing.Any]):
        """
        Create a reader executor task for a stream.
        """

        return self.loop.create_task(self.executor_wrapper(background_reader, stream, self.loop, callback))

    ANSI_ESCAPE_CODE = re.compile(r'\x1b\[\??(\d*)(?:([ABCDEFGJKSThilmnsu])|;(\d+)([fH]))')

    def clean_bytes(self, line: bytes) -> str:
        """
        Cleans a byte sequence of shell directives and decodes it.
        """

        text = line.decode('utf-8').replace('\r', '').strip('\n')

        def sub(group: typing.Match[str]):
            return group.group(0) if group.group(2) == 'm' and not self.escape_ansi else ''

        return self.ANSI_ESCAPE_CODE.sub(sub, text).replace("``", "`\u200b`").strip('\n')

    async def stdout_handler(self, line: bytes):
        """
        Handler for this class for stdout.
        """

        await self.queue.put(self.clean_bytes(line))

    async def stderr_handler(self, line: bytes):
        """
        Handler for this class for stderr.
        """

        await self.queue.put(self.clean_bytes(b'[stderr] ' + line))

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.process.kill()
        self.process.terminate()
        self.close_code = self.process.wait(timeout=0.5)

    def __aiter__(self):
        return self

    async def __anext__(self):
        last_output = time.perf_counter()

        while not self.closed or not self.queue.empty():
            try:
                item = await asyncio.wait_for(self.queue.get(), timeout=1)
            except asyncio.TimeoutError as exception:
                if time.perf_counter() - last_output >= self.timeout:
                    pass
            else:
                last_output = time.perf_counter()
                return item

        raise StopAsyncIteration()


import discord
from discord.ext import commands
import logging

logger: logging.Logger = logging.getLogger(__name__)

class log(commands.Cog):
    def __init__(self, bot):
        self.bot = bot




    @commands.Cog.listener()
    async def on_ready(self):
        logger.info(f'[ system ] udo in enabled bot: {self.bot.user.name}')




# -*- coding: utf-8 -*-

"""
jishaku.repl.inspections
~~~~~~~~~~~~~~~~~~~~~~~~~

Inspections performable on Python objects.

:copyright: (c) 2021 Devon (Gorialis) R
:license: MIT, see LICENSE for more details.

"""

import collections
import functools
import inspect
import os

INSPECTIONS = []
MethodWrapperType = type((1).__le__)
WrapperDescriptorType = type(int.__le__)


def add_inspection(name):
    """
    Add a Jishaku object inspection
    """

    # create the real decorator
    def inspection_inner(func):
        """
        Jishaku inspection decorator
        """

        # pylint: disable=inconsistent-return-statements

        # create an encapsulated version of the inspection that swallows exceptions
        @functools.wraps(func)
        def encapsulated(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except (TypeError, AttributeError, ValueError, OSError):
                return

        INSPECTIONS.append((name, encapsulated))
        return func

    return inspection_inner


def all_inspections(obj):
    """
    Generator to iterate all current Jishaku inspections.
    """

    for name, callback in INSPECTIONS:
        result = callback(obj)
        if result:
            yield name, result


def class_name(obj):
    """
    Get the name of an object, including the module name if available.
    """

    name = obj.__name__
    module = getattr(obj, "__module__")

    if module:
        name = f"{module}.{name}"
    return name


# pylint: disable=missing-docstring
# pylint: disable=inconsistent-return-statements


@add_inspection("Type")
def type_inspection(obj):
    return type(obj).__name__


@add_inspection("Object ID")
def id_inspection(obj):
    return hex(id(obj))


@add_inspection("Length")
def len_inspection(obj):
    return len(obj)


@add_inspection("MRO")
def mro_inspection(obj):
    if not inspect.isclass(obj):
        return

    return ", ".join(class_name(x) for x in inspect.getmro(obj))


@add_inspection("Type MRO")
def type_mro_inspection(obj):
    obj_type = type(obj)
    if obj_type in (type, object):
        return

    return ", ".join(class_name(x) for x in inspect.getmro(obj_type))


@add_inspection("Subclasses")
def subclass_inspection(obj):
    if isinstance(obj, type) and hasattr(obj, "__subclasses__"):
        subclasses = type.__subclasses__(obj)
    else:
        return

    output = ", ".join(class_name(x) for x in subclasses[0:5])

    if len(subclasses) > 5:
        output += ", ..."

    return output


@add_inspection("Module Name")
def module_inspection(obj):
    return inspect.getmodule(obj).__name__


@add_inspection("File Location")
def file_loc_inspection(obj):
    file_loc = inspect.getfile(obj)
    cwd = os.getcwd()
    if file_loc.startswith(cwd):
        file_loc = "." + file_loc[len(cwd) :]
    return file_loc


@add_inspection("Line Span")
def line_span_inspection(obj):
    source_lines, source_offset = inspect.getsourcelines(obj)
    return f"{source_offset}-{source_offset + len(source_lines)}"


@add_inspection("Signature")
def sig_inspection(obj):
    return inspect.signature(obj)


@add_inspection("Content Types")
def content_type_inspection(obj):
    if not isinstance(obj, (tuple, list, set)):
        return

    total = len(obj)
    types = collections.Counter(type(x) for x in obj)

    output = ", ".join(f"{x.__name__} ({y*100/total:.1f}\uFF05)" for x, y in types.most_common(3))
    if len(types) > 3:
        output += ", ..."

    return output


POSSIBLE_OPS = {
    "<": "lt",
    "<=": "le",
    "==": "eq",
    "!=": "ne",
    ">": "gt",
    ">=": "ge",
    "+": "add",
    "-": "sub",
    "*": "mul",
    "@": "matmul",
    "/": "truediv",
    "//": "floordiv",
    "%": "mod",
    "**": "pow",
    "<<": "lshift",
    ">>": "rshift",
    "&": "and",
    "^": "xor",
    "|": "or",
}


def check_not_slot(obj, attr):
    """
    Check that a given attribute isn't just an open slot for subclasses
    """

    if isinstance(getattr(obj, attr, None), MethodWrapperType):
        return not isinstance(getattr(type(obj), attr, None), WrapperDescriptorType)

    return not isinstance(getattr(obj, attr, None), WrapperDescriptorType)


@add_inspection("Operations")
def compat_operation_inspection(obj):
    this_dict = dir(obj)
    operations = []

    for operation, member in POSSIBLE_OPS.items():
        if f"__{member}__" in this_dict and check_not_slot(obj, f"__{member}__"):
            operations.append(operation)
        elif f"__r{member}__" in this_dict and check_not_slot(obj, f"r__{member}__"):
            operations.append(operation)

        if f"__i{member}__" in this_dict and check_not_slot(obj, f"i__{member}__"):
            operations.append(f"{operation}=")

    return " ".join(operations)




import asyncio
import functools
import typing


def executor_function(sync_function: typing.Callable):
    """A decorator that wraps a sync function in an executor, changing it into an async function.
    This allows processing functions to be wrapped and used immediately as an async function.
    Examples
    ---------
    Pushing processing with the Python Imaging Library into an executor:
    .. code-block:: python3
        from io import BytesIO
        from PIL import Image
        from jishaku.functools import executor_function
        @executor_function
        def color_processing(color: discord.Color):
            with Image.new('RGB', (64, 64), color.to_rgb()) as im:
                buff = BytesIO()
                im.save(buff, 'png')
            buff.seek(0)
            return buff
        @bot.command()
        async def color(ctx: commands.Context, color: discord.Color=None):
            color = color or ctx.author.color
            buff = await color_processing(color=color)
            await ctx.send(file=discord.File(fp=buff, filename='color.png'))
    """

    @functools.wraps(sync_function)
    async def sync_wrapper(*args, **kwargs):
        """
        Asynchronous function that wraps a sync function with an executor.
        """

        loop = asyncio.get_event_loop()
        internal_function = functools.partial(sync_function, *args, **kwargs)
        return await loop.run_in_executor(None, internal_function)

    return sync_wrapper


class AsyncSender:
    """
    Storage and control flow class that allows prettier value sending to async iterators.
    Example
    --------
    .. code:: python3
        async def foo():
            print("foo yielding 1")
            x = yield 1
            print(f"foo received {x}")
            yield 3
        async for send, result in AsyncSender(foo()):
            print(f"asyncsender received {result}")
            send(2)
    Produces:
    .. code::
        foo yielding 1
        asyncsender received 1
        foo received 2
        asyncsender received 3
    """

    __slots__ = ("iterator", "send_value")

    def __init__(self, iterator):
        self.iterator = iterator
        self.send_value = None

    def __aiter__(self):
        return self._internal(self.iterator.__aiter__())

    async def _internal(self, base):
        try:
            while True:
                # Send the last value to the iterator
                value = await base.asend(self.send_value)
                # Reset it incase one is not sent next iteration
                self.send_value = None
                # Yield sender and iterator value
                yield self.set_send_value, value
        except StopAsyncIteration:
            pass

    def set_send_value(self, value):
        """
        Sets the next value to be sent to the iterator.
        This is provided by iteration of this class and should
        not be called directly.
        """

        self.send_value = value



import asyncio

import discord
from discord.ext import commands

from udo.core.shim.paginator_base import EMOJI_DEFAULT


class PaginatorInterface:  # pylint: disable=too-many-instance-attributes
    """
    A message and reaction based interface for paginators.
    This allows users to interactively navigate the pages of a Paginator, and supports live output.
    An example of how to use this with a standard Paginator:
    .. code:: python3
        from discord.ext import commands
        from jishaku.paginators import PaginatorInterface
        # In a command somewhere...
            # Paginators need to have a reduced max_size to accommodate the extra text added by the interface.
            paginator = commands.Paginator(max_size=1900)
            # Populate the paginator with some information
            for line in range(100):
                paginator.add_line(f"Line {line + 1}")
            # Create and send the interface.
            # The 'owner' field determines who can interact with this interface. If it's None, anyone can use it.
            interface = PaginatorInterface(ctx.bot, paginator, owner=ctx.author)
            await interface.send_to(ctx)
            # send_to creates a task and returns control flow.
            # It will raise if the interface can't be created, e.g., if there's no reaction permission in the channel.
            # Once the interface has been sent, line additions have to be done asynchronously, so the interface can be updated.
            await interface.add_line("My, the Earth sure is full of things!")
            # You can also check if it's closed using the 'closed' property.
            if not interface.closed:
                await interface.add_line("I'm still here!")
    """

    def __init__(self, bot: commands.Bot, paginator: commands.Paginator, **kwargs):
        if not isinstance(paginator, commands.Paginator):
            raise TypeError("paginator must be a commands.Paginator instance")

        self._display_page = 0

        self.bot = bot

        self.message = None
        self.paginator = paginator

        self.owner = kwargs.pop("owner", None)
        self.emojis = kwargs.pop("emoji", EMOJI_DEFAULT)
        self.timeout = kwargs.pop("timeout", 7200)
        self.delete_message = kwargs.pop("delete_message", False)

        self.sent_page_reactions = False

        self.task: asyncio.Task = None
        self.send_lock: asyncio.Event = asyncio.Event()

        self.close_exception: Exception = None

        if self.page_size > self.max_page_size:
            raise ValueError(f"Paginator passed has too large of a page size for this interface. " f"({self.page_size} > {self.max_page_size})")

    @property
    def pages(self):
        """
        Returns the paginator's pages without prematurely closing the active page.
        """
        # protected access has to be permitted here to not close the paginator's pages

        # pylint: disable=protected-access
        paginator_pages = list(self.paginator._pages)
        if len(self.paginator._current_page) > 1:
            paginator_pages.append("\n".join(self.paginator._current_page) + "\n" + (self.paginator.suffix or ""))
        # pylint: enable=protected-access

        return paginator_pages

    @property
    def page_count(self):
        """
        Returns the page count of the internal paginator.
        """

        return len(self.pages)

    @property
    def display_page(self):
        """
        Returns the current page the paginator interface is on.
        """

        self._display_page = max(0, min(self.page_count - 1, self._display_page))
        return self._display_page

    @display_page.setter
    def display_page(self, value):
        """
        Sets the current page the paginator is on. Automatically pushes values inbounds.
        """

        self._display_page = max(0, min(self.page_count - 1, value))

    max_page_size = 2000

    @property
    def page_size(self) -> int:
        """
        A property that returns how large a page is, calculated from the paginator properties.
        If this exceeds `max_page_size`, an exception is raised upon instantiation.
        """
        page_count = self.page_count
        return self.paginator.max_size + len(f"\nPage {page_count}/{page_count}")

    @property
    def send_kwargs(self) -> dict:
        """
        A property that returns the kwargs forwarded to send/edit when updating the page.
        As this must be compatible with both `discord.TextChannel.send` and `discord.Message.edit`,
        it should be a dict containing 'content', 'embed' or both.
        """

        display_page = self.display_page
        page_num = f"\nPage {display_page + 1}/{self.page_count}"
        content = self.pages[display_page] + page_num
        return {"content": content}

    async def add_line(self, *args, **kwargs):
        """
        A proxy function that allows this PaginatorInterface to remain locked to the last page
        if it is already on it.
        """

        display_page = self.display_page
        page_count = self.page_count

        self.paginator.add_line(*args, **kwargs)

        new_page_count = self.page_count

        if display_page + 1 == page_count:
            # To keep position fixed on the end, update position to new last page and update message.
            self._display_page = new_page_count

        # Unconditionally set send lock to try and guarantee page updates on unfocused pages
        self.send_lock.set()

    async def send_to(self, destination: discord.abc.Messageable):
        """
        Sends a message to the given destination with this interface.
        This automatically creates the response task for you.
        """

        self.message = await destination.send(**self.send_kwargs)

        # add the close reaction
        await self.message.add_reaction(self.emojis.close)

        self.send_lock.set()

        if self.task:
            self.task.cancel()

        self.task = self.bot.loop.create_task(self.wait_loop())

        # if there is more than one page, and the reactions haven't been sent yet, send navigation emotes
        if not self.sent_page_reactions and self.page_count > 1:
            await self.send_all_reactions()

        return self

    async def send_all_reactions(self):
        """
        Sends all reactions for this paginator, if any are missing.
        This method is generally for internal use only.
        """

        for emoji in filter(None, self.emojis):
            try:
                await self.message.add_reaction(emoji)
            except discord.NotFound:
                # the paginator has probably already been closed
                break
        self.sent_page_reactions = True

    @property
    def closed(self):
        """
        Is this interface closed?
        """

        if not self.task:
            return False
        return self.task.done()

    async def send_lock_delayed(self):
        """
        A coroutine that returns 1 second after the send lock has been released
        This helps reduce release spam that hits rate limits quickly
        """

        gathered = await self.send_lock.wait()
        self.send_lock.clear()
        await asyncio.sleep(1)
        return gathered

    async def wait_loop(self):  # pylint: disable=too-many-branches, too-many-statements
        """
        Waits on a loop for reactions to the message. This should not be called manually - it is handled by `send_to`.
        """

        start, back, forward, end, close = self.emojis

        def check(payload: discord.RawReactionActionEvent):
            """
            Checks if this reaction is related to the paginator interface.
            """

            owner_check = not self.owner or payload.user_id == self.owner.id

            emoji = payload.emoji
            if isinstance(emoji, discord.PartialEmoji) and emoji.is_unicode_emoji():
                emoji = emoji.name

            tests = (
                owner_check,
                payload.message_id == self.message.id,
                emoji,
                emoji in self.emojis,
                payload.user_id != self.bot.user.id,
            )

            return all(tests)

        task_list = [
            self.bot.loop.create_task(coro)
            for coro in {
                self.bot.wait_for("raw_reaction_add", check=check),
                self.bot.wait_for("raw_reaction_remove", check=check),
                self.send_lock_delayed(),
            }
        ]

        try:  # pylint: disable=too-many-nested-blocks
            last_kwargs = None

            while not self.bot.is_closed():
                done, _ = await asyncio.wait(task_list, timeout=self.timeout, return_when=asyncio.FIRST_COMPLETED)

                if not done:
                    raise asyncio.TimeoutError

                for task in done:
                    task_list.remove(task)
                    payload = task.result()

                    if isinstance(payload, discord.RawReactionActionEvent):
                        emoji = payload.emoji
                        if isinstance(emoji, discord.PartialEmoji) and emoji.is_unicode_emoji():
                            emoji = emoji.name

                        if emoji == close:
                            await self.message.delete()
                            return

                        if emoji == start:
                            self._display_page = 0
                        elif emoji == end:
                            self._display_page = self.page_count - 1
                        elif emoji == back:
                            self._display_page -= 1
                        elif emoji == forward:
                            self._display_page += 1

                        if payload.event_type == "REACTION_ADD":
                            task_list.append(self.bot.loop.create_task(self.bot.wait_for("raw_reaction_add", check=check)))
                        elif payload.event_type == "REACTION_REMOVE":
                            task_list.append(self.bot.loop.create_task(self.bot.wait_for("raw_reaction_remove", check=check)))
                    else:
                        # Send lock was released
                        task_list.append(self.bot.loop.create_task(self.send_lock_delayed()))

                if not self.sent_page_reactions and self.page_count > 1:
                    self.bot.loop.create_task(self.send_all_reactions())
                    self.sent_page_reactions = True  # don't spawn any more tasks

                if self.send_kwargs != last_kwargs:
                    try:
                        await self.message.edit(**self.send_kwargs)
                    except discord.NotFound:
                        # something terrible has happened
                        return

                    last_kwargs = self.send_kwargs

        except (asyncio.CancelledError, asyncio.TimeoutError) as exception:
            self.close_exception = exception

            if self.bot.is_closed():
                # Can't do anything about the messages, so just close out to avoid noisy error
                return

            if self.delete_message:
                return await self.message.delete()

            for emoji in filter(None, self.emojis):
                try:
                    await self.message.remove_reaction(emoji, self.bot.user)
                except (discord.Forbidden, discord.NotFound):
                    pass

        finally:
            for task in task_list:
                task.cancel()


class PaginatorEmbedInterface(PaginatorInterface):
    """
    A subclass of :class:`PaginatorInterface` that encloses content in an Embed.
    """

    def __init__(self, *args, **kwargs):
        self._embed = kwargs.pop("embed", None) or discord.Embed()
        super().__init__(*args, **kwargs)

    @property
    def send_kwargs(self) -> dict:
        display_page = self.display_page
        self._embed.description = self.pages[display_page]
        self._embed.set_footer(text=f"Page {display_page + 1}/{self.page_count}")
        return {"embed": self._embed}

    max_page_size = 2048

    @property
    def page_size(self) -> int:
        return self.paginator.max_size


# -*- coding: utf-8 -*-

"""
jishaku.repl.compilation
~~~~~~~~~~~~~~~~~~~~~~~~

Constants, functions and classes related to classifying, compiling and executing Python code.

:copyright: (c) 2021 Devon (Gorialis) R
:license: MIT, see LICENSE for more details.

"""

import ast
import asyncio
import inspect
import linecache

import import_expression

from udo.functools import AsyncSender
from udo.repl.scope import Scope
from udo.repl.walkers import KeywordTransformer

CORO_CODE = f"""
async def _repl_coroutine({{0}}):
    import asyncio
    from importlib import import_module as {import_expression.constants.IMPORTER}

    import aiohttp
    import discord
    from discord.ext import commands

    try:
        import jishaku
    except ImportError:
        jishaku = None  # keep working even if in panic recovery mode

    try:
        pass
    finally:
        _async_executor.scope.globals.update(locals())
"""


def wrap_code(code: str, args: str = "") -> ast.Module:
    """
    Compiles Python code into an async function or generator,
    and automatically adds return if the function body is a single evaluation.
    Also adds inline import expression support.
    """

    user_code = import_expression.parse(code, mode="exec")
    mod = import_expression.parse(CORO_CODE.format(args), mode="exec")

    definition = mod.body[-1]  # async def ...:
    assert isinstance(definition, ast.AsyncFunctionDef)

    try_block = definition.body[-1]  # try:
    assert isinstance(try_block, ast.Try)

    try_block.body.extend(user_code.body)

    ast.fix_missing_locations(mod)

    KeywordTransformer().generic_visit(try_block)

    last_expr = try_block.body[-1]

    # if the last part isn't an expression, ignore it
    if not isinstance(last_expr, ast.Expr):
        return mod

    # if the last expression is not a yield
    if not isinstance(last_expr.value, ast.Yield):
        # copy the value of the expression into a yield
        yield_stmt = ast.Yield(last_expr.value)
        ast.copy_location(yield_stmt, last_expr)
        # place the yield into its own expression
        yield_expr = ast.Expr(yield_stmt)
        ast.copy_location(yield_expr, last_expr)

        # place the yield where the original expression was
        try_block.body[-1] = yield_expr

    return mod


class AsyncCodeExecutor:  # pylint: disable=too-few-public-methods
    """
    Executes/evaluates Python code inside of an async function or generator.

    Example
    -------

    .. code:: python3

        total = 0

        # prints 1, 2 and 3
        async for x in AsyncCodeExecutor('yield 1; yield 2; yield 3'):
            total += x
            print(x)

        # prints 6
        print(total)
    """

    __slots__ = ("args", "arg_names", "code", "loop", "scope", "source")

    def __init__(
        self,
        code: str,
        scope: Scope = None,
        arg_dict: dict = None,
        loop: asyncio.BaseEventLoop = None,
    ):
        self.args = [self]
        self.arg_names = ["_async_executor"]

        if arg_dict:
            for key, value in arg_dict.items():
                self.arg_names.append(key)
                self.args.append(value)

        self.source = code
        self.code = wrap_code(code, args=", ".join(self.arg_names))
        self.scope = scope or Scope()
        self.loop = loop or asyncio.get_event_loop()

    def __aiter__(self):
        exec(compile(self.code, "<repl>", "exec"), self.scope.globals, self.scope.locals)  # pylint: disable=exec-used
        func_def = self.scope.locals.get("_repl_coroutine") or self.scope.globals["_repl_coroutine"]

        return self.traverse(func_def)

    async def traverse(self, func):
        """
        Traverses an async function or generator, yielding each result.

        This function is private. The class should be used as an iterator instead of using this method.
        """

        try:
            if inspect.isasyncgenfunction(func):
                async for send, result in AsyncSender(func(*self.args)):
                    send((yield result))
            else:
                yield await func(*self.args)
        except Exception:  # pylint: disable=broad-except
            # Falsely populate the linecache to make the REPL line appear in tracebacks
            linecache.cache["<repl>"] = (
                len(self.source),  # Source length
                None,  # Time modified (None bypasses expunge)
                [line + "\n" for line in self.source.splitlines()],  # Line list
                "<repl>",  # 'True' filename
            )

            raise



import logging
from ._version import __version__
from .core.udo import udo
from .core.jeju_log import *
from os.path import isfile
import os
import json

Directory = os.path.dirname(os.path.realpath(__file__))

logger: logging.Logger = logging.getLogger(__name__)

if not isfile(f'{Directory}/core/data.json'):
    sdd = {
            "udo": {
                "tag": "udo",
                "value": "**udo project for py-cord**\n```py\npip install -U udo\n```\n```py\nbot.load_extension('udo')\n# or\nawait bot.load_extension('udo')\n```"
            } 
        }
    with open(f"{Directory}/core/data.json", "w",encoding="utf-8-sig") as json_file:
        json.dump(sdd,json_file,ensure_ascii = False, indent=4)
else:
    logger.info("[ udo ] The initial setup was skipped because the file was already created.")

async def setup(bot):
    await bot.add_cog(log(bot))
    await bot.add_cog(udo(bot))


# -*- coding: utf-8 -*-

"""
jishaku.repl.repl_builtins
~~~~~~~~~~~~~~~~~~~~~~~~~~~

Builtin functions and variables within Jishaku REPL contexts.

:copyright: (c) 2021 Devon (Gorialis) R
:license: MIT, see LICENSE for more details.

"""

import aiohttp
import discord
from discord.ext import commands


async def http_get_bytes(*args, **kwargs) -> bytes:
    """
    Performs a HTTP GET request against a URL, returning the response payload as bytes.

    The arguments to pass are the same as :func:`aiohttp.ClientSession.get`.
    """

    async with aiohttp.ClientSession() as session:
        async with session.get(*args, **kwargs) as response:
            response.raise_for_status()

            return await response.read()


async def http_get_json(*args, **kwargs) -> dict:
    """
    Performs a HTTP GET request against a URL,
    returning the response payload as a dictionary of the response payload interpreted as JSON.

    The arguments to pass are the same as :func:`aiohttp.ClientSession.get`.
    """

    async with aiohttp.ClientSession() as session:
        async with session.get(*args, **kwargs) as response:
            response.raise_for_status()

            return await response.json()


async def http_post_bytes(*args, **kwargs) -> bytes:
    """
    Performs a HTTP POST request against a URL, returning the response payload as bytes.

    The arguments to pass are the same as :func:`aiohttp.ClientSession.post`.
    """

    async with aiohttp.ClientSession() as session:
        async with session.post(*args, **kwargs) as response:
            response.raise_for_status()

            return await response.read()


async def http_post_json(*args, **kwargs) -> dict:
    """
    Performs a HTTP POST request against a URL,
    returning the response payload as a dictionary of the response payload interpreted as JSON.

    The arguments to pass are the same as :func:`aiohttp.ClientSession.post`.
    """

    async with aiohttp.ClientSession() as session:
        async with session.post(*args, **kwargs) as response:
            response.raise_for_status()

            return await response.json()


def get_var_dict_from_ctx(ctx: commands.Context, prefix: str = "_"):
    """
    Returns the dict to be used in REPL for a given Context.
    """

    raw_var_dict = {
        "author": ctx.author,
        "bot": ctx.bot,
        "channel": ctx.channel,
        "ctx": ctx,
        "find": discord.utils.find,
        "get": discord.utils.get,
        "guild": ctx.guild,
        "http_get_bytes": http_get_bytes,
        "http_get_json": http_get_json,
        "http_post_bytes": http_post_bytes,
        "http_post_json": http_post_json,
        "message": ctx.message,
        "msg": ctx.message,
    }

    return {f"{prefix}{k}": v for k, v in raw_var_dict.items()}




import collections
import typing

__all__ = ('Codeblock', 'codeblock_converter')


class Codeblock(typing.NamedTuple):
    """
    Represents a parsed codeblock from codeblock_converter
    """

    language: typing.Optional[str]
    content: str


def codeblock_converter(argument: str) -> Codeblock:
    """
    A converter that strips codeblock markdown if it exists.
    Returns a namedtuple of (language, content).
    :attr:`Codeblock.language` is an empty string if no language was given with this codeblock.
    It is ``None`` if the input was not a complete codeblock.
    """
    if not argument.startswith('`'):
        return Codeblock(None, argument)

    # keep a small buffer of the last chars we've seen
    last: typing.Deque[str] = collections.deque(maxlen=3)
    backticks = 0
    in_language = False
    in_code = False
    language: typing.List[str] = []
    code: typing.List[str] = []

    for char in argument:
        if char == '`' and not in_code and not in_language:
            backticks += 1  # to help keep track of closing backticks
        if last and last[-1] == '`' and char != '`' or in_code and ''.join(last) != '`' * backticks:
            in_code = True
            code.append(char)
        if char == '\n':  # \n delimits language and code
            in_language = False
            in_code = True
        # we're not seeing a newline yet but we also passed the opening ```
        elif ''.join(last) == '`' * 3 and char != '`':
            in_language = True
            language.append(char)
        elif in_language:  # we're in the language after the first non-backtick character
            if char != '\n':
                language.append(char)

        last.append(char)

    if not code and not language:
        code[:] = last

    return Codeblock(''.join(language), ''.join(code[len(language):-backticks]))


import os
import pathlib
import re
import subprocess

from setuptools import setup
import setuptools
from udo._version import __version__

STABLE_STAGE = True

ROOT = pathlib.Path(__file__).parent

EXTRA_REQUIRES = {}


for feature in (ROOT / "requirements").glob("*.txt"):
    with open(feature, "r", encoding="utf-8") as f:
        EXTRA_REQUIRES[feature.with_suffix("").name] = f.read().splitlines()
REQUIREMENTS = EXTRA_REQUIRES.pop("_")

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="udo.py", ## 소문자 영단어
    version=__version__, ##
    license="MIT",
    author="blket.dev", ## ex) Sunkyeong Lee
    author_email=None, ##
    description="discord is discord.py bot debug tools\nA discord.py extension including useful tools for bot development and debugging.", ##
    long_description=long_description,
    extras_require=EXTRA_REQUIRES,
    long_description_content_type="text/markdown",
    url=None, ##
    packages=["udo","udo.core","udo.repl","udo.core.shim","udo.core.work"],

    keywords="udo, discord.py , discord, cog, repl, extension, jishaku",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: AsyncIO",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Communications :: Chat",
        "Topic :: Internet",
        "Topic :: Software Development :: Debuggers",
        "Topic :: Software Development :: Testing",
        "Topic :: Utilities",
    ],
    python_requires='>=3.9',
)



import pathlib
import typing

import discord
import importlib.metadata
from braceexpand import braceexpand
from discord.ext import commands

from .core.work.types import BotT, ContextA

__all__ = ('find_extensions_in', 'resolve_extensions', 'package_version', 'ExtensionConverter')


if typing.TYPE_CHECKING:
    UnbalancedBracesError = ValueError
else:
    from braceexpand import UnbalancedBracesError


if typing.TYPE_CHECKING or discord.version_info >= (2, 0, 0):
    _ExtensionConverterBase = commands.Converter[typing.List[str]]
else:
    _ExtensionConverterBase = commands.Converter



def find_extensions_in(path: typing.Union[str, pathlib.Path]) -> list:
    """
    Tries to find things that look like bot extensions in a directory.
    """

    if not isinstance(path, pathlib.Path):
        path = pathlib.Path(path)

    if not path.is_dir():
        return []

    extension_names = []

    # Find extensions directly in this folder
    for subpath in path.glob("*.py"):
        parts = subpath.with_suffix("").parts
        if parts[0] == ".":
            parts = parts[1:]

        extension_names.append(".".join(parts))

    # Find extensions as subfolder modules
    for subpath in path.glob("*/__init__.py"):
        parts = subpath.parent.parts
        if parts[0] == ".":
            parts = parts[1:]

        extension_names.append(".".join(parts))

    return extension_names


def resolve_extensions(bot: commands.Bot, name: str) -> list:
    """
    Tries to resolve extension queries into a list of extension names.
    """

    exts = []
    for ext in braceexpand(name):
        if ext.endswith(".*"):
            module_parts = ext[:-2].split(".")
            path = pathlib.Path(*module_parts)
            exts.extend(find_extensions_in(path))
        elif ext == "~":
            exts.extend(bot.extensions)
        else:
            exts.append(ext)

    return exts

def package_version(package_name: str) -> typing.Optional[str]:
    """
    Returns package version as a string, or None if it couldn't be found.
    """

    try:
        return importlib.metadata.version(package_name)
    except importlib.metadata.PackageNotFoundError:
        return None


class ExtensionConverter(commands.Converter):  # pylint: disable=too-few-public-methods
    """
    A converter interface for resolve_extensions to match extensions from users.
    """

    async def convert(self, ctx: commands.Context, argument) -> list:
        try:
            return resolve_extensions(ctx.bot, argument)
        except UnbalancedBracesError as exc:
            raise commands.BadArgument(str(exc))




import typing

import discord

# emoji settings, this sets what emoji are used for PaginatorInterface

_Emoji = typing.Union[str, discord.PartialEmoji, discord.Emoji]


class EmojiSettings(typing.NamedTuple):
    """
    Emoji settings, this sets what emoji are used for PaginatorInterface
    """

    start: _Emoji
    back: _Emoji
    forward: _Emoji
    end: _Emoji
    close: _Emoji


EMOJI_DEFAULT = EmojiSettings(
    start="\N{BLACK LEFT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}",
    back="\N{BLACK LEFT-POINTING TRIANGLE}",
    forward="\N{BLACK RIGHT-POINTING TRIANGLE}",
    end="\N{BLACK RIGHT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}",
    close="\N{BLACK SQUARE FOR STOP}",
)



import dataclasses
import inspect
import os
import typing

import discord

from udo.core.work.types import ContextA

ENABLED_SYMBOLS = ("true", "t", "yes", "y", "on", "1")
DISABLED_SYMBOLS = ("false", "f", "no", "n", "off", "0")


FlagHandler = typing.Optional[typing.Callable[['FlagMeta'], typing.Any]]


@dataclasses.dataclass
class Flag:
    """
    Dataclass that represents a Jishaku flag state. Only for internal use.
    """

    name: str
    flag_type: type
    default: FlagHandler = None
    handler: FlagHandler = None
    override: typing.Any = None

    def resolve_raw(self, flags: 'FlagMeta'):  # pylint: disable=too-many-return-statements
        """
        Receive the intrinsic value for this flag, before optionally being processed by the handler.
        """

        # Manual override, ignore environment in this case
        if self.override is not None:
            return self.override

        # Resolve from environment
        env_value = os.getenv(f"JISHAKU_{self.name}", "").strip()

        if env_value:
            if self.flag_type is bool:
                if env_value.lower() in ENABLED_SYMBOLS:
                    return True
                if env_value.lower() in DISABLED_SYMBOLS:
                    return False
            else:
                return self.flag_type(env_value)

        # Fallback if no resolvation from environment
        if self.default is not None:
            if inspect.isfunction(self.default):
                return self.default(flags)

            return self.default

        return self.flag_type()

    def resolve(self, flags: 'FlagMeta'):
        """
        Resolve this flag. Only for internal use.
        Applies the handler when there is one.
        """

        value = self.resolve_raw(flags)

        if self.handler:
            return self.handler(value)  # type: ignore

        return value


class FlagMeta(type):
    """
    Metaclass for Flags.
    This handles the Just-In-Time evaluation of flags, allowing them to be overridden during execution.
    """

    def __new__(
        cls,
        name: str,
        base: typing.Tuple[typing.Type[typing.Any]],
        attrs: typing.Dict[str, typing.Any]
    ):
        attrs['flag_map'] = {}

        for flag_name, flag_type in attrs['__annotations__'].items():
            default: typing.Union[
                FlagHandler,
                typing.Tuple[
                    FlagHandler,  # default
                    FlagHandler,  # handler
                ],
            ] = attrs.pop(flag_name, None)
            handler: FlagHandler = None

            if isinstance(default, tuple):
                default, handler = default

            attrs['flag_map'][flag_name] = Flag(flag_name, flag_type, default, handler)

        return super(FlagMeta, cls).__new__(cls, name, base, attrs)

    def __getattr__(cls, name: str):
        cls.flag_map: typing.Dict[str, Flag]

        if hasattr(cls, 'flag_map') and name in cls.flag_map:
            return cls.flag_map[name].resolve(cls)

        return super().__getattribute__(name)

    def __setattr__(cls, name: str, value: typing.Any):
        if name in cls.flag_map:
            flag = cls.flag_map[name]

            if not isinstance(value, flag.flag_type):
                raise ValueError(f"Attempted to set flag {name} to type {type(value).__name__} (should be {flag.flag_type.__name__})")

            flag.override = value
        else:
            super().__setattr__(name, value)


class Flags(metaclass=FlagMeta):  # pylint: disable=too-few-public-methods
    """
    The flags for Jishaku.
    You can override these either through your environment, e.g.:
        export JISHAKU_HIDE=1
    Or you can override them programmatically:
        jishaku.Flags.HIDE = True
    """

    # Flag to indicate the Jishaku base command group should be hidden
    HIDE: bool

    # Flag to indicate that retention mode for REPL should be enabled by default
    RETAIN: bool

    # Flag to indicate that meta variables in REPL should not be prefixed with an underscore
    NO_UNDERSCORE: bool

    # The scope prefix, i.e. the prefix that appears before Jishaku's builtin variables in REPL sessions.
    # It is recommended that you set this programatically.
    SCOPE_PREFIX: str = lambda flags: '' if flags.NO_UNDERSCORE else '_'  # type: ignore

    # Flag to indicate whether to always use paginators over relying on Discord's file preview
    FORCE_PAGINATOR: bool

    # Flag to indicate verbose error tracebacks should be sent to the invoking channel as opposed to via direct message.
    # ALWAYS_DM_TRACEBACK takes precedence over this
    NO_DM_TRACEBACK: bool

    # Flag to indicate all errors, even minor ones like SyntaxErrors, should be sent via direct message.
    ALWAYS_DM_TRACEBACK: bool

    @classmethod
    def traceback_destination(cls, message: discord.Message) -> typing.Optional[discord.abc.Messageable]:
        """
        Determine what 'default' location to send tracebacks to
        When None, the caller should decide
        """

        if cls.ALWAYS_DM_TRACEBACK:
            return message.author

        if cls.NO_DM_TRACEBACK:
            return message.channel

        # Otherwise let the caller decide
        return None

    # Flag to indicate usage of braille J in shutdown command
    USE_BRAILLE_J: bool

    # Flag to indicate whether ANSI support should always be enabled
    # USE_ANSI_NEVER takes precedence over this
    USE_ANSI_ALWAYS: bool

    # Flag to indicate whether ANSI support should always be disabled
    USE_ANSI_NEVER: bool

    @classmethod
    def use_ansi(cls, ctx: ContextA) -> bool:
        """
        Determine whether to use ANSI support from flags and context
        """

        if cls.USE_ANSI_NEVER:
            return False

        if cls.USE_ANSI_ALWAYS:
            return True

        return not ctx.author.is_on_mobile() if isinstance(ctx.author, discord.Member) and ctx.bot.intents.presences else True